package ru.dbotthepony.mc.otm.client.screen.panels.button

import com.mojang.blaze3d.platform.InputConstants
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.block.entity.RedstoneSetting
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.Widgets18
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.client.screen.panels.EditablePanel
import ru.dbotthepony.mc.otm.client.screen.panels.FramePanel
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.math.RelativeSide
import ru.dbotthepony.mc.otm.menu.input.BooleanInputWithFeedback
import ru.dbotthepony.mc.otm.menu.input.EnergyConfigPlayerInput
import ru.dbotthepony.mc.otm.menu.input.IPlayerInputWithFeedback
import ru.dbotthepony.mc.otm.menu.input.ItemConfigPlayerInput
import java.util.function.Predicate

private fun <S : MatteryScreen<*>> makeRedstoneSettingButton(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	control: IPlayerInputWithFeedback<RedstoneSetting>
): LargeEnumRectangleButtonPanel<S, RedstoneSetting> {
	return object : LargeEnumRectangleButtonPanel<S, RedstoneSetting>(
		screen, parent, x = x, y = y,
		defaultValue = RedstoneSetting.LOW, enum = RedstoneSetting::class.java,
		prop = control,
	) {
		override var isDisabled: Boolean
			get() = !control.test(minecraft.player)
			set(value) {}

		init {
			add(RedstoneSetting.IGNORED, tooltip = RedstoneSetting.IGNORED.description, skinElement = Widgets18.REDSTONE_IGNORED)
			add(RedstoneSetting.LOW, tooltip = RedstoneSetting.LOW.description, skinElement = Widgets18.REDSTONE_LOW)
			add(RedstoneSetting.HIGH, tooltip = RedstoneSetting.HIGH.description, skinElement = Widgets18.REDSTONE_HIGH)
		}
	}
}

private fun <S : MatteryScreen<*>> makeItemModeButton(screen: S, parent: FramePanel<S>, input: ItemConfigPlayerInput.Piece, side: RelativeSide): LargeEnumRectangleButtonPanel<S, MatteryDeviceBlockEntity.ItemHandlerMode> {
	val button = LargeEnumRectangleButtonPanel(screen, parent, enum = MatteryDeviceBlockEntity.ItemHandlerMode::class.java, prop = input.input, defaultValue = input.default)

	for (v in MatteryDeviceBlockEntity.ItemHandlerMode.values()) {
		button.add(v, skinElement = Widgets18.ITEMS_CONTROLS[side]!![v]!!, tooltip = TranslatableComponent(v.translationKey))
	}

	button.finish()
	button.predicate = Predicate { input.isAllowed(it) }

	return button
}

private fun <S : MatteryScreen<*>> makeEnergyModeButton(screen: S, parent: FramePanel<S>, input: EnergyConfigPlayerInput.Piece, side: RelativeSide): LargeEnumRectangleButtonPanel<S, FlowDirection> {
	val button = LargeEnumRectangleButtonPanel(screen, parent, enum = FlowDirection::class.java, prop = input.input, defaultValue = input.default)

	for (v in FlowDirection.values()) {
		button.add(v, skinElement = Widgets18.CONTROLS[side]!![v]!!, tooltip = TranslatableComponent(v.translationKey))
	}

	button.finish()

	return button
}

private fun moveButtons(
	front: EditablePanel<*>,
	back: EditablePanel<*>,
	left: EditablePanel<*>,
	right: EditablePanel<*>,
	top: EditablePanel<*>,
	bottom: EditablePanel<*>,
) {
	top.tooltip = TranslatableComponent("otm.gui.sides.top")
	bottom.tooltip = TranslatableComponent("otm.gui.sides.bottom")
	back.tooltip = TranslatableComponent("otm.gui.sides.back")
	front.tooltip = TranslatableComponent("otm.gui.sides.front")
	left.tooltip = TranslatableComponent("otm.gui.sides.left")
	right.tooltip = TranslatableComponent("otm.gui.sides.right")

	top.x = 30f
	top.y = 14f

	right.x = 30f - 20f
	right.y = 14f + 20f

	left.x = 30f + 20f
	left.y = 14f + 20f

	front.x = 30f
	front.y = 14f + 20f

	bottom.x = 30f
	bottom.y = 14f + 42f

	back.x = 30f - 20f
	back.y = 14f + 42f
}

@Suppress("name_shadowing")
private fun pullPush(frame: FramePanel<*>, pull: BooleanInputWithFeedback, push: BooleanInputWithFeedback) {
	if (pull.test(minecraft.player)) {
		val pull = LargeBooleanRectangleButtonPanel(
			frame.screen,
			frame,
			skinElementActive = Widgets18.PULL,
			skinElementInactive = Widgets18.PULL_DISABLED,
			prop = pull,
		)

		pull.tooltip = TranslatableComponent("otm.gui.side_mode.pull")

		pull.x = 30f - 20f
		pull.y = 14f
	}

	if (push.test(minecraft.player)) {
		val push = LargeBooleanRectangleButtonPanel(
			frame.screen,
			frame,
			skinElementActive = Widgets18.PUSH,
			skinElementInactive = Widgets18.PUSH_DISABLED,
			prop = push,
		)

		push.tooltip = TranslatableComponent("otm.gui.side_mode.push")

		push.x = 30f + 20f
		push.y = 14f
	}
}

private fun <S : MatteryScreen<*>> makeItemHandlerControlPanel(
	screen: S,
	inputs: ItemConfigPlayerInput
): FramePanel<S> {
	val frame = object : FramePanel<S>(screen, 78f, 80f, TranslatableComponent("otm.gui.sides.item_config")) {
		override fun tickInner() {
			super.tickInner()

			if (!isEverFocused()) {
				remove()
			}
		}
	}

	val front = makeItemModeButton(screen, frame, inputs.pieces[RelativeSide.FRONT]!!, RelativeSide.FRONT)
	val back = makeItemModeButton(screen, frame, inputs.pieces[RelativeSide.BACK]!!, RelativeSide.BACK)
	val left = makeItemModeButton(screen, frame, inputs.pieces[RelativeSide.LEFT]!!, RelativeSide.LEFT)
	val right = makeItemModeButton(screen, frame, inputs.pieces[RelativeSide.RIGHT]!!, RelativeSide.RIGHT)
	val top = makeItemModeButton(screen, frame, inputs.pieces[RelativeSide.TOP]!!, RelativeSide.TOP)
	val bottom = makeItemModeButton(screen, frame, inputs.pieces[RelativeSide.BOTTOM]!!, RelativeSide.BOTTOM)

	pullPush(frame, inputs.pull, inputs.push)
	moveButtons(front, back, left, right, top, bottom)
	screen.addPanel(frame)
	frame.requestFocus()

	return frame
}

private fun <S : MatteryScreen<*>> makeEnergyConfigPanel(
	screen: S,
	inputs: EnergyConfigPlayerInput
): FramePanel<S> {
	val frame = object : FramePanel<S>(screen, 78f, 80f, TranslatableComponent("otm.gui.sides.energy_config")) {
		override fun tickInner() {
			super.tickInner()

			if (!isEverFocused()) {
				remove()
			}
		}
	}

	val front = makeEnergyModeButton(screen, frame, inputs.pieces[RelativeSide.FRONT]!!, RelativeSide.FRONT).also { it.predicate = Predicate { inputs.possibleModes.isSupertype(it) } }
	val back = makeEnergyModeButton(screen, frame, inputs.pieces[RelativeSide.BACK]!!, RelativeSide.BACK).also { it.predicate = Predicate { inputs.possibleModes.isSupertype(it) } }
	val left = makeEnergyModeButton(screen, frame, inputs.pieces[RelativeSide.LEFT]!!, RelativeSide.LEFT).also { it.predicate = Predicate { inputs.possibleModes.isSupertype(it) } }
	val right = makeEnergyModeButton(screen, frame, inputs.pieces[RelativeSide.RIGHT]!!, RelativeSide.RIGHT).also { it.predicate = Predicate { inputs.possibleModes.isSupertype(it) } }
	val top = makeEnergyModeButton(screen, frame, inputs.pieces[RelativeSide.TOP]!!, RelativeSide.TOP).also { it.predicate = Predicate { inputs.possibleModes.isSupertype(it) } }
	val bottom = makeEnergyModeButton(screen, frame, inputs.pieces[RelativeSide.BOTTOM]!!, RelativeSide.BOTTOM).also { it.predicate = Predicate { inputs.possibleModes.isSupertype(it) } }

	pullPush(frame, inputs.pull, inputs.push)
	moveButtons(front, back, left, right, top, bottom)
	screen.addPanel(frame)
	frame.requestFocus()

	return frame
}

class DeviceControls<out S : MatteryScreen<*>>(
	screen: S,
	parent: FramePanel<S>,
	extra: Iterable<EditablePanel<S>> = listOf(),
	val redstoneConfig: IPlayerInputWithFeedback<RedstoneSetting>? = null,
	val itemConfig: ItemConfigPlayerInput? = null,
	val energyConfig: EnergyConfigPlayerInput? = null,
) : EditablePanel<S>(screen, parent, x = parent.width + 3f, height = 0f, width = 0f) {
	val itemConfigButton: LargeRectangleButtonPanel<S>?
	val energyConfigButton: LargeRectangleButtonPanel<S>?
	val redstoneControlsButton: LargeEnumRectangleButtonPanel<S, RedstoneSetting>?
	private var nextY = 0f

	fun <P : EditablePanel<@UnsafeVariance S>> addButton(button: P): P {
		button.parent = this
		button.x = 0f
		button.y = nextY
		nextY += button.height + 2f
		height = nextY - 2f
		width = button.width.coerceAtLeast(width)
		return button
	}

	init {
		for (button in extra) {
			addButton(button)
		}

		if (redstoneConfig != null) {
			redstoneControlsButton = addButton(makeRedstoneSettingButton(screen, this, control = redstoneConfig))
		} else {
			redstoneControlsButton = null
		}

		if (itemConfig != null) {
			itemConfigButton = addButton(object : LargeRectangleButtonPanel<S>(screen, this@DeviceControls, skinElement = Widgets18.ITEMS_CONFIGURATION) {
				init {
					tooltip = TranslatableComponent("otm.gui.sides.item_config")
				}

				override fun onClick(mouseButton: Int) {
					if (mouseButton == InputConstants.MOUSE_BUTTON_LEFT) {
						val frame = makeItemHandlerControlPanel(screen, itemConfig)

						frame.x = absoluteX + width / 2f - frame.width / 2f
						frame.y = absoluteY + height + 8f
					}
				}
			})
		} else {
			itemConfigButton = null
		}

		if (energyConfig != null) {
			energyConfigButton = addButton(object : LargeRectangleButtonPanel<S>(screen, this@DeviceControls, y = nextY, skinElement = Widgets18.ENERGY_CONFIGURATION) {
				init {
					tooltip = TranslatableComponent("otm.gui.sides.energy_config")
				}

				override fun onClick(mouseButton: Int) {
					if (mouseButton == InputConstants.MOUSE_BUTTON_LEFT) {
						val frame = makeEnergyConfigPanel(screen, energyConfig)

						frame.x = absoluteX + width / 2f - frame.width / 2f
						frame.y = absoluteY + height + 8f
					}
				}
			})
		} else {
			energyConfigButton = null
		}
	}

	override fun tickInner() {
		super.tickInner()
		x = (parent?.width ?: 0f) + 3f
		y = 0f
	}
}

fun <S : MatteryScreen<*>> makeDeviceControls(
	screen: S,
	parent: FramePanel<S>,
	extra: Iterable<EditablePanel<S>> = listOf(),
	redstoneConfig: IPlayerInputWithFeedback<RedstoneSetting>? = null,
	itemConfig: ItemConfigPlayerInput? = null,
	energyConfig: EnergyConfigPlayerInput? = null,
): DeviceControls<S> {
	return DeviceControls(screen, parent, extra = extra, redstoneConfig = redstoneConfig, itemConfig = itemConfig, energyConfig = energyConfig)
}
